import sys
import argparse
from mlgenotype import rfgenotype

def init_argparse() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        usage="%(prog)s [OPTION] [FILE]...",
        description="Train a random forest model and write it and its accuracy stats out to the file system."
    )
    parser.add_argument(
        "-v", "--version", action="version",
        version = f"{parser.prog} version 1.0.0"
    )
    parser.add_argument('-s', '--simdata', metavar='tab-delimited file of features/genotypes to train on', required=True)
    parser.add_argument('-t', '--testdata', metavar='tab-delimited file of validation features/genotypes on which to measure the models accuracy (optional).')
    parser.add_argument('-o', '--outputfilebase', default='model', metavar='filename base for output files. Model will be written to <filebase>.model, accuracy statistics to <filebase>.stats')

    return parser

def main() -> None:
   parser = init_argparse()
   args = parser.parse_args()

   # reads in simulated dataset that will be used to fit our RF model:
   simdatafile = args.simdata
   print("Training with simulated data file: " + simdatafile)

   # if data for validating the fitted model exists, read that in:
   valdatafile = args.testdata
   testsim_x = None
   if valdatafile is not None:
       print("Checking fitted models accuracy with data in file: " + valdatafile)
       testsim_x, testsim_y, testcolumnnames, testuniquegenos = rfgenotype.read_data_file(valdatafile, shuffle=False)

   # report where output will go:
   outputbase = args.outputfilebase
   print("Will write model to " + outputbase + ".rf.model, model accuracy stats to " + outputbase + ".rf.stats, and feature importance values to " + outputbase + ".rf.importance.txt")

   trainsim_x, trainsim_y, traincolumnnames, trainuniquegenos = rfgenotype.read_data_file(simdatafile)

   # now optimize parameters using GridSearchCV (saving GSCV's best estimator model fit on all the data with the best params)
   grid_sim = rfgenotype.rfgrid(trainsim_x, trainsim_y)
   best_params_sim = grid_sim.best_params_  # finds best parameters
   rf_model =  grid_sim.best_estimator_  # model fit with all data and best hyperparameters
   best_params_score = grid_sim.best_score_ # mean CV score of model fit with all data and best hyperparameters

   # save the optimized model to <outputbase>.rf.model:
   rfgenotype.save_model_to_file(rf_model, outputbase + ".rf.model")

   # write out feature importance values:
   importances = rf_model.feature_importances_
   importancefile = outputbase + ".rf.importance.txt"
   with open(importancefile, 'w') as impwriter:
      for imp in importances:
         impwriter.write(str(imp) + '\n')

   statsfile = outputbase + ".rf.stats"
   with open(statsfile, 'w') as statswriter:
      statswriter.write("RF model best hyperparameters:\nBest number of estimators: " + str(best_params_sim['n_estimators']) + "\nBest maximum depth: " + str(best_params_sim['max_depth']) + "\n")
      statswriter.write("RF model mean CV score (all data, best hyperparameters):\n" + str(best_params_score) + "\n")

      # testing simulated model with validation samples, if they were provided
      if testsim_x is not None:
         test_preds = rf_model.predict(testsim_x)  # predicting x variables from validation set using model
         test_actual = testsim_y
         print("Accuracy of fitted RF model for held out validation set:")
         test_accuracy = rfgenotype.accuracy_score(test_actual, test_preds)
         statswriter.write(str(test_accuracy) + "\n")

if __name__ == "__main__":
    main()



import sys
import argparse
from mlgenotype import rfgenotype

def init_argparse() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        usage="%(prog)s [OPTION] [FILE]...",
        description="Train a random forest model and use it to predict genotypes on real data."
    )
    parser.add_argument(
        "-v", "--version", action="version",
        version = f"{parser.prog} version 1.0.0"
    )
    parser.add_argument('-s', '--simdata', default='./TrainingData.txt', metavar='tab-delimited file of features/genotypes to train on')
    parser.add_argument('-r', '--realdata', default='', metavar='tab-delimited file of features to run fitted predictions on')
    return parser

def main() -> None:
   parser = init_argparse()
   args = parser.parse_args()

   # reads in simulated dataset that will be used to create model for real data
   simdatafile = args.simdata
   print("Simulated data file: " + simdatafile)
   simfeatures, simgenotypes, columnnames, uniquegenos = rfgenotype.read_data_file(simdatafile)
   trainsim_x, testsim_x, trainsim_y, testsim_y = rfgenotype.separate_test_data(simfeatures, simgenotypes)

   # now optimize parameters using GridSearchCV (saving GSCV's best estimator model fit on all the data with the best params)
   grid_sim = rfgenotype.rfgrid(trainsim_x, trainsim_y)
   model_grid =  grid_sim.best_estimator_  # model from grid

   # write out feature importance values:
   importances = model_grid.feature_importances_
   with open('feature_importance_values.txt', 'w') as writer:
      for imp in importances:
         writer.write(str(imp) + '\n')

   # then separately train a model with best params on all the training data:
   best_params_sim = grid_sim.best_params_  # finds best parameters
   cal_score_output_sim = rfgenotype.best_rf_model_and_score(trainsim_x.to_numpy(), trainsim_y.to_numpy(), best_params_sim['n_estimators'], best_params_sim['max_depth'])
   model_sim = cal_score_output_sim[0]  # model from simulated data
   model_sim_scores = cal_score_output_sim[1]
   
   # testing simulated model with 100 test points that were taken out beforehand
   test_preds = model_sim.predict(testsim_x)  # predicting x variables from test set using model
   test_actual = testsim_y
   print("Estimated CV scores for best parameters:")
   print(model_sim_scores)
   print("Accuracy of model recalculated using calculate_score function for predictions on 100 held out test samples:")
   print(rfgenotype.accuracy_score(test_actual, test_preds))

   # testing simulated model with 100 test points that were taken out beforehand
   test_preds = model_grid.predict(testsim_x)  # predicting x variables from test set using model
   print("Accuracy of best estimator model from GridSearchCV on 100 held out test samples:")
   print(rfgenotype.accuracy_score(test_actual, test_preds))

   # reads in real data
   realdatafile = args.realdata
   if realdatafile != '':
      realfeatures, realgenos, realcolumns, realuniquegenos = rfgenotype.read_data_file(realdatafile, shuffle=False)
      y_pred_grid = model_grid.predict(realfeatures.to_numpy())  # finds genotype probability for predictions using model
   
      y_prob_grid = model_grid.predict_proba(realfeatures.to_numpy()) # finds genotype probability for predictions using GridSearchCV

      # creates textfiles to put all predictions/probability predictions in 
      with open('genotype_output_grid.txt', 'w') as writer:
         for index, genotype in enumerate(y_pred_grid):
            writer.write(str(index+1) + ': ' + genotype + '\n')
      with open('genotype_output_grid_probs.txt', 'w') as writer:
         for i in range(len(y_prob_grid)):
            output = y_prob_grid[i]
            writer.write(str(output) + '\n')

if __name__ == "__main__":
    main()



import pandas as pd
from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import cross_val_score
from sklearn.model_selection import StratifiedKFold
from sklearn.model_selection import GridSearchCV
from sklearn.metrics import accuracy_score
import math
import pickle

# reads a file of features and target values, tab-delimited, with one sample per line
def read_data_file(filename_path, shuffle=True):
 
    # reads in dataset
    df = pd.read_csv(filename_path, sep='\t')
    
    # randomly shuffles dataset
    if shuffle:
        df = df.sample(frac=1)
    
    # drops 'Genotype' column from original dataframe, saves 'x' variables
    df_x = df.drop(['Genotype'], axis=1)

    # represents actual mutated genotype for each observation, saves 'y' variables
    df_y = df['Genotype'] 

    # grabs coverage for each chromosomal section (note--this includes 'Genotype' column)
    df_column_names = df.columns
    
    # stores the 6 mutation genotypes in a list
    genotype = df.Genotype.unique()

    return df_x, df_y, df_column_names, genotype

# splits data set, reserving some samples to use as a test set: note that samples should have been shuffled in the read_data_file routine for this to be unbiased separation
def separate_test_data(df_x, df_y, holdoutnum=100):
    # takes 100 random samples out and stores it
    df_variables_test = df_x[:holdoutnum]
    df_output_genotype_test = df_y[:holdoutnum]

    # keeps rest of variables not taken out to train model
    df_variables_train = df_x[holdoutnum:]
    df_output_genotype_train = df_y[holdoutnum:]
    
    return df_variables_train, df_variables_test, df_output_genotype_train, df_output_genotype_test

# finds best parameters using GridSearchCV
def rfgrid(df_x, df_y):
    limit = int(math.log2(len(df_x.index)))

    n_trees = [100, 200, 300, 400, 500]
    max_dep = [2, 3, 4, 5, 6, 7, 8, 9, 10]

    if (limit in max_dep) == False:
        for m in max_dep:
            if limit < m:
                max_dep.insert(max_dep.index(m), limit)
                break

    if max_dep[-1] < limit:
        max_dep.append(limit)
    
    parameters = {'n_estimators':n_trees, 'max_depth':max_dep}
    
    rf_grid = GridSearchCV(RandomForestClassifier(random_state=2), param_grid=parameters,
                          return_train_score=True, cv=5)
    rf_grid.fit(df_x, df_y)
    
    return rf_grid


# calculates cross validation score of an RF model with default hyper parameters
# ALSO returns model created. BUT can also just use .best_estimator from GridSearchCV
# tutorial referenced:
### https://www.youtube.com/watch?v=gJo0uNL-5Qw
def best_rf_model_and_score(featuredf, genos, best_n_trees, best_max_depth):

    model = RandomForestClassifier(n_estimators=best_n_trees, max_depth=best_max_depth,
                                       random_state=2)
    model.fit(featuredf, genos)

    cross_score = cross_val_score(model, featuredf, genos, cv=5)
    
    return model, cross_score

def save_model_to_file(fittedmodel, filename):

    pickle.dump(fittedmodel, open(filename, 'wb'))

def load_model_from_file(filename):

    loaded_model = pickle.load(open(filename, 'rb'))

    return loaded_model




import sys
import argparse
from mlgenotype import rfgenotype

def init_argparse() -> argparse.ArgumentParser:
     parser = argparse.ArgumentParser(
         usage="%(prog)s [OPTION] [FILE]...",
         description="Generate data for the manuscript."
     )
     parser.add_argument(
         "-v", "--version", action="version",
         version = f"{parser.prog} version 1.0.0"
     )
     parser.add_argument('-r', '--train', default='training_data.txt', metavar='file of sample k-mer counts with one sample per line and tab-delimited feature values for each sample. File should contain a header, and the final field of each line should be the genotype')
     parser.add_argument('-t', '--test', default='training_data.txt', metavar='file of sample k-mer counts with one sample per line and tab-delimited feature values for each sample. File should contain a header, and the final field of each line should be the genotype')
     parser.add_argument('-o', '--outdir', default='rfresults', metavar='directory for output files. Subdirectory models will have saved models, and subdirectory predictions will have the predictions of those models')
     parser.add_argument('-p', '--prefix', default='rfresults', metavar='prefix for output filenames')
 
     return parser

def main() -> None:

    parser = init_argparse()
    args = parser.parse_args()
 
    # Number of training samples to use:
    ntrainvals = [110, 220, 441, 882, 1764, 2646]
    #ntrainvals = [5292]
 
    # reads in simulated dataset that will be used to train our RF models, shuffling:
    traindatafile = args.train
    trainsim_x, trainsim_y, traincolumnnames, trainuniquegenos = rfgenotype.read_data_file(traindatafile, shuffle=True)
 
    # reads in simulated dataset that will be used to test our RF models, without shuffling:
    testdatafile = args.test
    testsim_x, testsim_y, testcolumnnames, testuniquegenos = rfgenotype.read_data_file(testdatafile, shuffle=False)
 
    for ntrain in ntrainvals:
        # pull out ntrain samples for training our model:
        holdout = len(trainsim_x) - ntrain
        train_x, test_x, train_y, test_y = rfgenotype.separate_test_data(trainsim_x, trainsim_y, holdoutnum=holdout)
        if len(train_x) != ntrain:
            print("train_x length is not as expected!")
            trainxlength = len(train_x)
            print("train_x length: " + str(trainxlength) + " holdout: " + str(holdout))
            sys.exit(1)
  
        # optimize hyperparameters using GridSearchCV:
        grid_sim = rfgenotype.rfgrid(train_x, train_y)
        best_params_sim = grid_sim.best_params_  # best hyperparameters
        rf_model =  grid_sim.best_estimator_  # model fit with all data and best hyperparameters
        best_params_score = grid_sim.best_score_ # mean CV score of model fit with all data and best hyperparameters
  
        # report where output will go:
        outputbase = args.prefix
        print("Writing model to " + outputbase + "." + str(ntrain) + ".rf.model, model accuracy stats to " + outputbase + "." + str(ntrain) + ".rf.stats, and feature importance values to " + outputbase + "." + str(ntrain) + ".rf.importance.txt in models directory" )
  
        # save the optimized model to <outputbase>.rf.model:
        rfgenotype.save_model_to_file(rf_model, args.outdir + "/models/" + outputbase + "." + str(ntrain) + ".rf.model")
  
        # write out feature importance values:
        importances = rf_model.feature_importances_
        importancefile = args.outdir + "/models/" + outputbase + "." + str(ntrain) + ".rf.importance.txt"
        with open(importancefile, 'w') as impwriter:
            for imp in importances:
                impwriter.write(str(imp) + '\n')
     
        statsfile = args.outdir + "/models/" + outputbase + "." + str(ntrain) + ".rf.stats"
        with open(statsfile, 'w') as statswriter:
            statswriter.write("RF model best hyperparameters:\nBest number of estimators: " + str(best_params_sim['n_estimators']) + "\nBest maximum depth: " + str(best_params_sim['max_depth']) + "\n")
            statswriter.write("RF model mean CV score (all data, best hyperparameters):\n" + str(best_params_score) + "\n")
  
        # now test this model with the appropriate test data and print results:
        test_preds = rf_model.predict(testsim_x.to_numpy())
        test_probs = rf_model.predict_proba(testsim_x.to_numpy())
        actual_genos = testsim_y
        modelpredfile = args.outdir + "/predictions/" + outputbase + "." + str(ntrain) + ".rf.genotypes.txt"
        with open(modelpredfile, 'w') as predwriter:
            for i in range(len(test_preds)):
                probs = test_probs[i]
                maxprob = max(probs) 
                predwriter.write(str(i + 1) + "\t" + str(test_preds[i]) + "\t" + str(actual_genos[i]) + "\t" + str(maxprob) + "\n")
   
if __name__ == "__main__":
    main()



import sys
import argparse
from mlgenotype import rfgenotype
import warnings
from sklearn.exceptions import DataConversionWarning

def init_argparse() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        usage="%(prog)s [OPTION] [FILE]...",
        description="Use a trained random forest model to predict genotypes and probabilities for a feature set."
    )
    parser.add_argument(
        "-v", "--version", action="version",
        version = f"{parser.prog} version 1.0.0"
    )
    parser.add_argument('-m', '--modelfile', metavar='pickle file with RF model to load, as created by RFModelTrain.py', required=True)
    parser.add_argument('-f', '--featurefile', metavar='tab-delimited file of validation features/genotypes on which to predict genotypes', required=True)
    parser.add_argument('-o', '--outputfilebase', default='model', metavar='filename base for output files. Genotypes will be written to <filebase>.rf.genos, probabilities to <filebase>.rf.probs')

    return parser

def main() -> None:
   parser = init_argparse()
   args = parser.parse_args()

   warnings.filterwarnings(action='ignore', category=UserWarning)

   # load the optimized model specified with the --modelfile argument:
   modelfile = args.modelfile
   rf_model = rfgenotype.load_model_from_file(modelfile)
   print("Using RF model loaded from file: " + modelfile)

   # reads in feature dataset that will be used for making predictions with the RF model:
   featurefile = args.featurefile
   print("Using RF model to make predictions from features in file: " + featurefile)

   # report where output will go:
   outputbase = args.outputfilebase
   print("Will write genotypes to " + outputbase + ".rf.genos, genotype probabilities to " + outputbase + ".rf.probs")

   data_x, data_y, columnnames, uniquegenos = rfgenotype.read_data_file(featurefile, shuffle=False)
   data_preds = rf_model.predict(data_x.to_numpy())  # predicting y variables from feature set using model
   data_probs = rf_model.predict_proba(data_x.to_numpy())  # predicted probabilities for different genotypes

   # creates textfiles to put all predictions/probability predictions in 
   with open(outputbase + '.rf.genos', 'w') as writer:
      for index, genotype in enumerate(data_preds):
         output = data_probs[index]
         maxprob = max(output)
         writer.write(str(index+1) + '\t' + genotype + '\t' + str(maxprob) + '\n')
   writer.close()
   #with open(outputbase + '.rf.probs', 'w') as writer:
      #for i in range(len(data_probs)):
         #output = data_probs[i]
         #maxprob = max(output)
         ##writer.write(str(output) + '\n')
         #writer.write(str(maxprob) + '\n')
   #writer.close()


if __name__ == "__main__":
    main()
